//
// guido2xml.cpp (main functions to convert guido -> xml)
//
// This file contains the implementation of the callback functions,
// wich will be called from the parser. There is also a dynamic stack
// implementation to store xml-tags and bracketed-information. Furthermore
// we have some basic file-operations to handle the output file-stream.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "guido.h"
#include "file.h"
#include "stack.h"
#include "taglist.h"
#include "guido2xml.h"
#include "defines.h"

// Globals!
// note: these definitions are mainly needed to get access in the callback functions
//       given from the guido-parser-kit
int      tab            = 0; // actual tabwith
int      tab_use        = 1; // indicates, if the tab should be used '= 1' or not used '= 0'
int      bracketed_flag = 0; // indicates, if a tag is bracketed '=1' or not '=0'

// the variable indicates that a tag is added (=1), this value should be in any 'tag-add-routine'.
// if tag_add = 0 the tags that are written with file_write will be stored temporarily
int      tag_add = 1; 

// here is the ability to store only the params given from the parser
int      tag_param = 0;                 // if param (=1), here is the notify
long int tmp_no    = 0;                 // temporary tag number, given from parser
char     tag_param_str[TAG_LENGTH_MAX]; // string of one tag param
char     tag_name[TAG_LENGTH_MAX];      // the actual xml tag-name

// the guido parse-function
int na_parseGMNFile(const char* filename) {
  int res;
  gd_init();
  res = gd_parse(filename, PARSE_MODE_ALL);
  if (res == -1) return -1;
  gd_exit();
  return res;
}

// the main parse function
int parseFile(const char* filename, const char* dtd_name) {
  int res;                  // result of actual function
  char* to_dot;             // pointer for string operations
  char xml_filename[1024];  // buffer to store the xml filename
  char dtd_filename[1024];  // buffer to store the dtd filename
  char buf[TAG_LENGTH_MAX]; // temporary buffer
  strcpy(xml_filename, filename);
  
  // find the first dot of actual filename and write the xml-ending behind,
  // if there's no dot then append the xml-ending
  if (!(to_dot = strchr(filename, '.'))) strcat(xml_filename, ".xml");
  else { 
    strncat(xml_filename, filename, to_dot-filename);       
    xml_filename[to_dot-filename] = '\0';
    strcat(xml_filename, ".xml");
  }
  
  printf("\nwriting '%s'.\n", xml_filename); fflush(stdout);

  reset_tmp_buf();

  if(!file_open(xml_filename)) {
      printf("ERROR: Can't create file '%s'\n", xml_filename);
      exit(0);
  }

  // write the beginning of xml-header
  file_write("<?xml version=\"", tag_add, tab_use, tab);
  sprintf(buf, XML_VERSION);
  file_write(buf, tag_add, tab_use, tab);
  file_write("\" encoding=\"", tag_add, tab_use, tab);
  sprintf(buf, XML_ENCODING);
  file_write(buf, tag_add, tab_use, tab);
  file_write("\"?>\n", tag_add, tab_use, tab);

  if (dtd_name != NULL) sprintf(dtd_filename, "%s", dtd_name);
  else sprintf(dtd_filename, XML_DEFAULT_SYSTEM);
  
  file_write("<!DOCTYPE ", tag_add, tab_use, tab);
  sprintf(buf, XML_DOCTYPE);
  file_write(buf, tag_add, tab_use, tab);
  sprintf(buf, " SYSTEM \"%s\">\n", dtd_filename);
  file_write(buf, tag_add, tab_use, tab);
  
  file_write("<", tag_add, tab_use, tab);
  sprintf(buf, XML_DOCTYPE);
  file_write(buf, tag_add, tab_use, tab);
  file_write(">\n", tag_add, tab_use, tab);

  tab += TAB_LENGTH;

  // now, parse the given file
  res = na_parseGMNFile(filename);  

  tab -= TAB_LENGTH;

  //write the ending of xml-header
  file_write("</", tag_add, tab_use, tab);
  sprintf(buf, XML_DOCTYPE);
  file_write(buf, tag_add, tab_use, tab);
  file_write(">\n", tag_add, tab_use, tab);

  file_close();
 
  if (res == 0) { 
      printf("job sucessfully done.\n\n");
      fflush(stdout);
  }

  return res;
}

// The following callback-functions return the right tag-argument
// according to it's type

char na_getTagArgType(int n) {
  return(gd_getTagArgType(n));
}

char* na_getTagArgName(int n) {
  return(gd_getTagArgName(n));
}

long int na_getTagArgInt(int n) {
  return(gd_getTagArgInt(n));
}

T_REAL na_getTagArgFloat(int n) {
  return(gd_getTagArgFloat(n));
}

char* na_getTagArgUnit(int n) {
  return(gd_getTagArgUnit(n));
}

char* na_getTagArgStr(int n) {
  return(gd_getTagArgStr(n));
}


// the following functions are the implementation of 
// callback-functions from the parser

// Beginnig of the segment
void na_segmInit() {
  tag_add = 1; // we have to add a tag
  file_write("<segment>\n", tag_add, tab_use, tab);
  tab += TAB_LENGTH; // makes better readability
}

// Ending of the segment
void na_segmExit() {
  tab -= TAB_LENGTH;
  file_write("</segment>\n", tag_add, tab_use, tab);
}

// begin of a sequence, alway store the id of the sequence
void na_seqInit() {
  char buf[TAG_LENGTH_MAX];

  tag_add = 1; // we have to add a tag
  
  sprintf(buf, "<sequence>\n");
  file_write(buf, tag_add, tab_use, tab);

  tab += TAB_LENGTH;
}

// the ending of a sequence
void na_seqExit() {
  char buf[TAG_LENGTH_MAX];

  tab -= TAB_LENGTH;

  sprintf(buf, "</sequence>\n");
  file_write(buf, tag_add, tab_use, tab);
}

// this function append one note
void na_seqAppendNote(int pc, int oct, int acc, long int durN, long int durD, int dots,
		      long int durPosN, long int durPosD){
  char buf[TAG_LENGTH_MAX];    
  char name[16]; // store the name of the note
  int  rest = 0; // if there is a rest, rest=1 else rest=0
  int  i = 0;

  tag_add = 1; // we have to add a tag

  // now check the temporary buffer
  // note: all known ambiguous tags have notes inside the parenthesis,
  //       so this is the most important buffer check
  tab = check_tmp_buf(tag_add, tab_use, tab);  

  // translate the note-defines to note-strings
  switch (pc) {
  case NOTE_C: sprintf(name, "c"); break;
  case NOTE_CIS: sprintf(name, "cis"); break;
  case NOTE_D: sprintf(name, "d"); break;
  case NOTE_DIS: sprintf(name, "dis"); break;
  case NOTE_E: sprintf(name, "e"); break;
  case NOTE_F: sprintf(name, "f"); break;
  case NOTE_FIS: sprintf(name, "fis"); break;
  case NOTE_G: sprintf(name, "g"); break;
  case NOTE_GIS: sprintf(name, "gis"); break;
  case NOTE_A: sprintf(name, "a"); break;
  case NOTE_AIS: sprintf(name, "ais"); break;
  case NOTE_H: sprintf(name, "h"); break;
  case REST: sprintf(name, "rest"); rest = 1; break;
  case EMPTY: sprintf(name, "empty"); break;    
  default: printf("\nERROR: unknown note found\n"); fflush(stdout); exit(0); break;
  }
    
  // decides if there is a rest, else it must be a note
  // there is also a check for using dots and accidentals (looks similar to the guido format)
  if (rest == 1) {
    sprintf(buf, "<rest duration=\"%li/%li", durN, durD);
    for (i = 0; i < dots; i++) strcat(buf, ".");
    strcat(buf, "\"/>\n");
    file_write(buf, tag_add, tab_use, tab);
  } else {
    sprintf(buf, "<note name=\"%s\" octave=\"%i\" duration=\"%li/%li", name , oct, durN, durD);
    for (i = 0; i < dots; i++) strcat(buf, ".");
    strcat(buf, "\"");
    if (abs(acc)) strcat(buf, " accidentals=\"");
    for (i = 0; i < abs(acc); i++)  {
      if (acc < 0) strcat(buf, "b"); // we use b instead &, because & is in xml not allowed
      else strcat(buf, "#");
    }
    if (abs(acc)) strcat(buf, "\"");
    strcat(buf, "/>\n");
    
    file_write(buf, tag_add, tab_use, tab);
  }
}

// beginning of a chord
void na_chordInit(long int durPosN, long int durPosD) {
  tag_add = 1; // we have to add a tag
  // perhaps there are some ambiguous tags with chords inside the parenthesis
  tab = check_tmp_buf(tag_add, tab_use, tab); 
  file_write("<chord>\n", tag_add, tab_use, tab);
  tab += TAB_LENGTH;
}

// append a chord
void na_seqAppendChord(int voices, long int durN, long int durD, int dots,
		       long int durPosN, long int durPosD) {
  tag_add = 1; // we have to add a tag
  // not really needed, buf safe
  tab = check_tmp_buf(tag_add, tab_use, tab); 
  tab -= TAB_LENGTH;
  file_write("</chord>\n", tag_add, tab_use, tab);
}


// all other tags we be handled here
void na_tagStart(char* s, long int no, long int tagno, long int durPosN, long int durPosD) {
  char buf[TAG_LENGTH_MAX];

  tag_add = 1; // we have to add a tag

  bracketed_flag = 0; // always beginn to set 'bracketed=no'

  // this is the main tag-check, where all known guido tags are translated to xml-tags
  // note: all xml-tags are pushed to the stack, but only the bracketed tags (with parenthesis) 
  //       are important to know, because the tagEnd-function have to know witch tag-ending
  //       belongs to to tag-beginning. There is also the bracketed-info stored in the stack,
  //       so the stack looks like: "tag-name", "x", "tag-name", "x", "tag-name", ...
  //       the x stands for "1" (bracketed=yes) or "0" (bracketed=no)  
  
  if (findGuidoTag(tag_bracketed, s) != NULL) {
    sprintf(tag_name, "/%s", findGuidoTag(tag_bracketed, s)); // xml-tag-end
    stack_push(tag_name);
    sprintf(tag_name, "%s", findGuidoTag(tag_bracketed, s));  // xml-tag-begin
    bracketed_flag = 1;
  }
  else if (findGuidoTag(tag_bracketed_ambiguous, s) != NULL) {
    sprintf(tag_name, "/%s", findGuidoTag(tag_bracketed_ambiguous, s)); // xml-tag-end
    stack_push(tag_name);
    sprintf(tag_name, "%s", findGuidoTag(tag_bracketed_ambiguous, s));  // xml-tag-begin
    bracketed_flag = 1;
    tag_add = 0; // we have to wait for adding this tag to see if it is one with parenthesis
    tmp_no = no; // store the actual tag-id
  }
  else if (findGuidoTag(tag_bracketed_none, s) != NULL) {
    sprintf(tag_name, "%s", findGuidoTag(tag_bracketed_none, s));
    stack_push(tag_name);
  }
  // take the rest as it is and check if ambiguous or not
  else {
    // remove backslash of the given tag
    // note: it is safe to do this operation, 
    //       because the life of the string s will be ended after this function exits
    printf("WARNING: unknown tag found: %s\n", s);
    s++;
    sprintf(tag_name, "/%s", s); // xml-tag-end
    stack_push(tag_name);
    sprintf(tag_name, "%s", s);  // xml-tag-begin
    bracketed_flag = 1;
    tag_add = 0; // we have to wait for adding this tag to see if it is one with parenthesis
    tmp_no = no; // store the actual tag-id
  }
  
  // the beginning of a xml-tag
  sprintf(buf, "<%s", tag_name); 
  file_write(buf, tag_add, tab_use, tab);

  // disable tabulator
  // note: this is needed, because now we start to add the params
  tab_use = 0;
  
  // if id is given then add it
  if (no > 0) {
    sprintf(buf, " id=\"%li\"", no);
    file_write(buf, tag_add, tab_use, tab);    
  }

  // now, push the bracketed-info to the stack
  if (bracketed_flag) {
    stack_push("1");
    tab += TAB_LENGTH;
  }
  else stack_push("0");
}

// this function is called at the end of a tag
void na_tagEnd(long int tagno, long int durPosN, long int durPosD, long int beforeDurPosN, long int beforeDurPosD) {
  char buf_tmp[TAG_LENGTH_MAX]; // temporary buffer to store the tags received from the stack
  char buffer[TAG_LENGTH_MAX]; // temporary buffer to store the end-tag for writing to file
  
  // first remove bracketed-info from stack, if there was a bracketed-tag then change tab-width back.
  // then writing the end of the tag (note: this will be done only if bracketed (with parenthesis))
  stack_pop(buf_tmp);
  if (strcmp(buf_tmp, "1") == 0) {
    tab -= TAB_LENGTH;
    
    // these special tags have no slash at the end, all others get one
    stack_front(buf_tmp);
    
    // here is the check of the ambiguous tags:
    // the parenthesis will be detected, if other tags (like notes) have been added (tag_add = 1)
    // if it was so, writing the prepared tags from the stack, but if not (tag_add = 0) there
    // is no origination to use the temporary buffer, so clean the content of this buffer
    if (!tag_add) {
      sprintf(buffer, "<");
            
      if (findGuidoTag(tag_bracketed_ambiguous, tag_name) != NULL) {
	sprintf(buf_tmp, "%s", findGuidoTag(tag_bracketed_ambiguous, tag_name)); 
	strcat(buffer, buf_tmp);
      }
      else {
	sprintf(buf_tmp, "%s", tag_name); strcat(buffer, buf_tmp);
      }

      if (tmp_no)    { sprintf(buf_tmp, "%li", tmp_no); strcat(buffer, buf_tmp); }
      if (tag_param) { sprintf(buf_tmp, "%s", tag_param_str); strcat(buffer, buf_tmp); }
      
      strcat(buffer, "/>\n");
      
      reset_tmp_buf(); // throw old tag-name away and use the new one
      tag_add = 1;
      stack_pop(buf_tmp);
    }
    else {
      tab = check_tmp_buf(tag_add, tab_use, tab); 
      stack_pop(buf_tmp);
      sprintf(buffer, "<%s>\n", buf_tmp);
    }

    file_write(buffer, tag_add, tab_use, tab);
  } else {
    stack_pop(buf_tmp); // throw stack top element away (in case of standalone tag)
  }
  tab_use = 1; // not really needed, buf safe!
}

// check if unit is given. If yes, add the unit to the buffer
void write_unit(char* buffer, int no) {
  char buf[TAG_LENGTH_MAX];

  // check given units
  if (strcmp(na_getTagArgUnit(no), "\0")) sprintf(buf, "%s\"", na_getTagArgUnit(no));
  else sprintf(buf, "\"");
  strcat(buffer, buf);
}

// choose the right param_type and put it into the given buffer
// note: &#x22; is used to have a definite string (&#x22; = ")
int write_param_type(char* buffer, int no, int name_found) {
  int is_unit = 0;

  // first, check if param is a unit
  if (strcmp(na_getTagArgUnit(no), "\0")) is_unit = 1;

  switch (na_getTagArgType(no)) {
  case 's':
    sprintf(buffer, "&#x22;%s&#x22;", na_getTagArgStr(no));    
    if (findXMLParam(tag_param_string, tag_name, no) != NULL && !name_found) {
      if (is_unit)
	sprintf(buffer, " %s=\"&#x22;%s&#x22;", findXMLParam(tag_param_unit, tag_name, no), na_getTagArgStr(no));
      else
	sprintf(buffer, " %s=\"&#x22;%s&#x22;", findXMLParam(tag_param_string, tag_name, no), na_getTagArgStr(no));
      return 1;
    }
    break;
  case 'i':
    sprintf(buffer, "%li", na_getTagArgInt(no));
    if (findXMLParam(tag_param_int, tag_name, no) != NULL && !name_found) {
      if (is_unit)
	sprintf(buffer, " %s=\"%li", findXMLParam(tag_param_unit, tag_name, no), na_getTagArgInt(no));
      else
	sprintf(buffer, " %s=\"%li", findXMLParam(tag_param_int, tag_name, no), na_getTagArgInt(no));
      return 1;
    }
    break;
  case 'f':
    sprintf(buffer, "%f", na_getTagArgFloat(no));
    if (findXMLParam(tag_param_float, tag_name, no) != NULL && !name_found) {
      if (is_unit)
	sprintf(buffer, " %s=\"%f", findXMLParam(tag_param_unit, tag_name, no), na_getTagArgFloat(no));
      else
	sprintf(buffer, " %s=\"%f", findXMLParam(tag_param_float, tag_name, no), na_getTagArgFloat(no));
      return 1;
    }
    break;
  default:
    buffer[0] = '\0';
    break;
  }
  return 0;
}

// write the param number 'no' to the xml-file
void write_param(char* buffer, int no) {
  char buf[TAG_LENGTH_MAX];
  int extra_param = 0, name_found = 0;

  // check if param_name was found
  // was soll das denn sein? alt : "\0"
  if (strcmp(na_getTagArgName(no), "")) name_found = 1;

  extra_param = write_param_type(buf, no, name_found);
  
  // if param-name is given: overwrite existing name stored in buffer...
  if (name_found) {
    sprintf(buffer, " %s=\"%s", na_getTagArgName(no), buf);
  }
  else { 
    if (extra_param) {
      file_write(buf, tag_add, tab_use, tab);
      buffer[0] = '\0';
    } else {
      file_write(buffer, tag_add, tab_use, tab); //...else write the given param from our stored list
      sprintf(buffer, "\"%s", buf);
    }
  }
  
  write_unit(buffer, no); // write the unit

  file_write(buffer, tag_add, tab_use, tab);  // write the whole param
}


// check additional param.
// note: if no tag-name is given, the write the default name: 'arg#' (# is the tag number)
void write_additional_param(char* buffer, int no) {
  char buf[TAG_LENGTH_MAX];
  int extra_param = 0, name_found = 0;

  // check if param_name was found
  if (strcmp(na_getTagArgName(no), "\0")) name_found = 1;

  extra_param = write_param_type(buf, no, name_found);
  
  if (name_found) {
    sprintf(buffer, " %s=\"%s", na_getTagArgName(no), buf); 
  } else {
    if (extra_param) {
      strcpy(buffer, buf);
    } else {
      sprintf(buffer, " arg%i=\"%s", no, buf);
    }
  }
  
  write_unit(buffer, no); // write the unit
  
  file_write(buffer, tag_add, tab_use, tab);  // write the whole param
}

// get number of rules (split-positions)
int get_rule_amount(char* tag_name) {
  int i = 0;
  while (findXMLParam(tag_param_rule, tag_name, ++i) != NULL);
  return i - 1;
}

// check if there a rule is given. If yes, then return the split-position
int check_param_rule(char* tag_name, int no) {
  if (findXMLParam(tag_param_rule, tag_name, no) != NULL) 
    return atoi(findXMLParam(tag_param_rule, tag_name, no)) - 1;
  return 0;
}

// handling of the tag params
// note: if params are given, check tag-name and add the given params
void na_tagAdd(long int tagno, int nargs) {
  char buf[TAG_LENGTH_MAX];
  buf[0] = 0;
  int i = 0, rule_pos = 0, rule_pos_act = 0, rule_no = 0;

  tag_param = 0; // this global variable is used to check if params are given
  
  // writing the params
  if (nargs >= 1) {
    tag_param = 1; // turn the tag_param flag on
    
    // now check all the split-points and take the biggest one that fits into the
    // given range
    if (nargs > 1) {
      rule_no = get_rule_amount(tag_name);
      for (i = 1; i <= rule_no; i++) {
      	rule_pos_act = check_param_rule(tag_name, i);
      	if (rule_pos_act > rule_pos  && rule_pos_act <= nargs)
	  rule_pos = rule_pos_act;
      }
    }
    
    // check all params on the three tag-classes and add to the param-list
    for (i = 1; i <= nargs; i++) {
      if (findXMLParam(tag_bracketed, tag_name, i + rule_pos) != NULL) {
	sprintf(buf, " %s=", findXMLParam(tag_bracketed, tag_name, i + rule_pos)); 
	write_param(buf, i);
      }
      else if (findXMLParam(tag_bracketed_ambiguous, tag_name, i + rule_pos) != NULL) {
	sprintf(buf, " %s=", findXMLParam(tag_bracketed_ambiguous, tag_name, i + rule_pos)); 
	write_param(buf, i);
      }
      else if (findXMLParam(tag_bracketed_none, tag_name, i + rule_pos) != NULL) {
	sprintf(buf, " %s=", findXMLParam(tag_bracketed_none, tag_name, i + rule_pos)); 
	write_param(buf, i);
      }
      else write_additional_param(buf, i); // this is for all param in unknown tags
    }    
  }
    
  strcpy(tag_param_str, buf); // get the actual param string

  // this code decides wich tag gets the slash at the end (only non-bracketed tags)
  if (bracketed_flag) sprintf(buf, ">\n");
  else sprintf(buf, "/>\n");
  
  file_write(buf, tag_add, tab_use, tab);

  tab_use = 1; // disable ignoring the tabs (now we have written the complete tag)

}

// if parser error is occured, print error-message
void na_parseError(long int lnr, long int cnr, const char *msg) {
  printf("\nERROR: %s (line %li, char %li)\n", msg, lnr, cnr);
}
