#ifndef TAGLIST_H
#define TAGLIST_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "defines.h"

//
// The guido->xml tag definitions.
// There are three classes of tag-lists and one rule tag-list.
// For details see below
//
// The following order is required:
//
// guido-tag xml-tag xml-param1 xml-param2 ... xml-paramN
//
// note: - blanks are only allowed between tags and params
//       - at the end of the list there must be a "#"
//       - if you have changed this file be sure to do first a 'make clean'
//         on unix systems and 'Rebuild All' on Visual C++.
//

//
// always bracketed tags
//
static const char* const tag_bracketed[] = { 
  "\\cresc          cresc          s r",
  "\\dim            dim            s r",
  "\\stacc          stacc",
  "\\accent         accent",
  "\\ten            ten",
  "\\marcato        marcato",
  "\\tie            tie            dx1 dx2 dy1 dy2 h",
  "\\slur           slur           dy dy1 dy2 dx1 dy1 dx2 dy2 dx3 dy3",
  "\\sl             sl             dy dy1 dy2 dx1 dy1 dx2 dy2 dx3 dy3",
  "\\accel          accel          s s2",
  "\\rit            rit            s s2",
  "\\grace          grace          i",
  "\\cue            cue            s",
  "\\beam           beam           dy dy1 dy2",
  "\\bm             bm             dy dy1 dy2",
  "\\trill          trill          i",
  "\\mord           mord           i",
  "\\turn           turn           i",
  "\\fermata        fermata",
  "\\oct            oct            i",
  "\\label          label          s",
  "\\repeatEnd      repeatEnd      i",

  // advanced guido
  "\\fingering      fingering      text dx dy fsize",
  "\\acc            acc            dx",
  "#"};

//
// ambiguous bracketed tags
// important: tags defined here are NOT well formed in XML, but
//            are allowed by the most XML parsers. The reasen of
//            that is because EMPTY defined tags are not included
//            in other entities.
//
// Example: fermata could look like
//            <fermata/> or <fermata>
//                             <note1 ...>
//                             <note2 ...>
//                          </fermata> 
//
static const char* const tag_bracketed_ambiguous[] = { 
  "#"};

//
// never bracketed tags
//
static const char* const tag_bracketed_none[] = { 
  "\\slurBegin      slurBegin      dy dy1 dy2 dx1 dy1 dx2 dy2 dx3 dy3",
  "\\slurEnd        slurEnd",
  "\\crescBegin     crescBegin",
  "\\crescEnd       crescEnd       s r",
  "\\dimBegin       dimBegin",
  "\\dimEnd         dimEnd         s r",
  "\\staccBegin     staccBegin",  // ?
  "\\staccEnd       staccEnd",    // ?
  "\\accentBegin    accentBegin", // ?
  "\\accentEnd      accentEnd",   // ?
  "\\tenBegin       tenBegin",    // ?
  "\\tenEnd         tenEnd",      // ?
  "\\marcatoBegin   marcatoBegin",// ? 
  "\\marcatoEnd     marcatoEnd",  // ?
  "\\tieBegin       tieBegin       dx1 dx2 dy1 dy2 h",
  "\\tieEnd         tieEnd",
  "\\accelBegin     accelBegin",
  "\\accelEnd       accelEnd       s s2",
  "\\ritBegin       ritBegin",
  "\\ritEnd         ritEnd         s s2",
  "\\repeatBegin    repeatBegin    i",
  "\\intens         intens         s r",
  "\\i              i              s r",
  "\\instr          instr          s s2",
  "\\tempo          tempo          s s2",
  "\\trem           trem           i",
  "\\beamsAuto      beamsAuto",
  "\\beamsOff       beamsOff",
  "\\stemsAuto      stemsAuto",
  "\\stemsUp        stemsUp        dy",
  "\\stemsDown      stemsDown      dy",
  "\\bar            bar            i",
  "\\doubleBar      doubleBar      i",
  "\\tactus         tactus",
  "\\text           text           s dy",
  "\\t              t              s dy",
  "\\clef           clef           s",
  "\\title          title          s pageformat dx dy",
  "\\composer       composer       s dy fsize",
  "\\mark           mark           s",
  "\\label          label          s",
  "\\meter          meter          s",
  "\\key            key",
  "\\staff          staff",

  // advanced guido
  "\\pageFormat     pageFormat     style lm tm bm rm",
  "\\autoBreak      autoBreak      system page",
  "\\systemFormat   systemFormat   dx",
  "\\staffFormat    staffFormat    style size",
  "\\restFormat     restFormat     posy dx dy",
  "\\newSystem      newSystem      dx dy",
  "\\space          space          size",
  "\\barFormat      barFormat      style",
  "\\dotFormat      dotFormat      dy",
  "\\newPage        newPage",
  "#"};

//
// Tag rule-list
//
// here, we can define a basic rule for the params of the tags
//
// the order: guido-tag xml-tag position1 position2 ... positionN
//
// the position represents where are the params should be splitted into
// N param lists. These split-points depend on the given amount of the
// params. 
//
// Example: 
//          The slur tag has the following params:
//          dy dy1 dy2 dx1 dy1 dx2 dy2 dx3 dy3
//          1  2   3   4   5   6   7   8   9
//
// the given split-points: 2 4
// The rule: - If slur has less than two params, then take 
//             param1 = dy
//           - If the param-number is more than one, then take 
//             param1 = dy1, param2 = dy2
//           - If the param-number is over 3 (>=4), then take
//             param1 = dx1, param2 = dy2, param3 = dx2, ...
//
// note: the guido- and XML-Tag are independent each other
static const char* const tag_param_rule[] = { 
  "\\slur           slur           2 4",
  "\\sl             sl             2 4",
  "\\slurBegin      slurBegin      2 4",
  "\\beam           beam           2",
  "\\bm             bm             2",
  "#"};

//
// Here we can define extra subsitute params depend
// on the argument type.
//
// The order: guido-tag xml-tag param1 param2 ... paramN
//
// note: the guido- and XML-Tag are independent each other
//
// replace params in case of string is detected
static const char* const tag_param_string[] = { 
  "\\key            key            s",
  "\\staff          staff          s",
  "#"};

// replace params in case of integer is detected
static const char* const tag_param_int[] = { 
  "\\key            key            i",
  "\\staff          staff          i",
  "#"};

// replace params in case of float is detected
static const char* const tag_param_float[] = { 
  "#"};

// replace params in case of unit is detected
static const char* const tag_param_unit[] = { 
  "#"};


// the functions to access to the string-lists
char*  findGuidoTag(const char*const*, char*);
char*  findGuidoParam(const char*const*, char*, int);
char*  findXMLTag(const char*const*, char*);
char*  findXMLParam(const char*const*, char*, int);

#endif
